#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2016 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #
"""
    pddoc.sphinx
    ~~~~~~~~~~~~~~~~~~~

    Allow puredata-formatted graphs to be included in Sphinx-generated
    documents inline.
"""

import posixpath
from os import path
from subprocess import Popen, PIPE
from hashlib import sha1

from docutils import nodes
from docutils.parsers.rst import directives
from docutils.statemachine import ViewList

import sphinx
from sphinx.errors import SphinxError
from sphinx.util.osutil import ensuredir, ENOENT, EPIPE, EINVAL
from sphinx.util.compat import Directive


class PdError(SphinxError):
    category = 'Pd error'


class PdNode(nodes.General, nodes.Inline, nodes.Element):
    pass


def figure_wrapper(directive, node, caption):
    figure_node = nodes.figure('', node)
    if 'align' in node:
        figure_node['align'] = node.attributes.pop('align')

    parsed = nodes.Element()
    directive.state.nested_parse(ViewList([caption], source=''),
                                 directive.content_offset, parsed)
    caption_node = nodes.caption(parsed[0].rawsource, '',
                                 *parsed[0].children)
    caption_node.source = parsed[0].source
    caption_node.line = parsed[0].line
    figure_node += caption_node
    return figure_node


def align_spec(argument):
    return directives.choice(argument, ('left', 'center', 'right'))


class PureData(Directive):
    """
    Directive to insert arbitrary ascii puredata markup.
    """
    has_content = True
    required_arguments = 0
    optional_arguments = 1
    final_argument_whitespace = False
    option_spec = {
        'alt': directives.unchanged,
        'align': align_spec,
        'caption': directives.unchanged,
    }

    def run(self):
        pd_ascii_code = '\n'.join(self.content)
        if not pd_ascii_code.strip():
            return [self.state_machine.reporter.warning(
                'Ignoring "pd" directive without content.',
                line=self.lineno)]

        node = PdNode()
        node['code'] = pd_ascii_code
        node['options'] = {}
        if 'alt' in self.options:
            node['alt'] = self.options['alt']
        if 'align' in self.options:
            node['align'] = self.options['align']

        caption = self.options.get('caption')
        if caption:
            node = figure_wrapper(self, node, caption)

        return [node]


def render_pd(self, code, options, format, prefix='pd'):
    """Render puredata code into a PNG or PDF output file."""
    pdgraph = options.get('pdascii', self.builder.config.pdgraph)
    hashkey = (code + str(options) +
               str(self.builder.config.pdgraph_args)).encode('utf-8')
    fname = '{0:s}-{1:s}.{2:s}'.format(prefix, sha1(hashkey).hexdigest(), format)
    relfn = posixpath.join(self.builder.imgpath, fname)
    outfn = path.join(self.builder.outdir, self.builder.imagedir, fname)
    tmp_txt = '{0:s}-{1:s}-pdascii.txt'.format(prefix, sha1(hashkey).hexdigest())
    tmp_path = path.join(self.builder.outdir, self.builder.imagedir, tmp_txt)

    ensuredir(path.dirname(outfn))
    ensuredir(path.dirname(tmp_path))

    if not path.exists(tmp_path):
        f = open(tmp_path, 'w')
        f.writelines(code)
        f.close()
    else:
        if path.isfile(outfn):
            return relfn, outfn

    if hasattr(self.builder, '_pdgraph_warned') and self.builder._pdgraph_warned.get(pdgraph):
        return None, None

    pd_args = [pdgraph]
    pd_args.extend(self.builder.config.pdgraph_args)
    pd_args.extend(['--format', format, tmp_path, outfn])

    try:
        p = Popen(pd_args, stdout=PIPE, stdin=PIPE, stderr=PIPE)
    except OSError as err:
        if err.errno != ENOENT:   # No such file or directory
            raise
        self.builder.warn('pdascii command cannot be run (needed for pd '
                          'output), check the pdgraph setting')
        if not hasattr(self.builder, '_pdgraph_warned'):
            self.builder._pdgraph_warned = {}
        self.builder._pdgraph_warned[pdgraph] = True
        return None, None

    try:
        # resulting in a broken pipe on communicate()
        stdout, stderr = p.communicate(code)
    except (OSError, IOError) as err:
        if err.errno not in (EPIPE, EINVAL):
            raise
        # in this case, read the standard output and standard error streams
        # directly, to get the error message(s)
        stdout, stderr = p.stdout.read(), p.stderr.read()
        p.wait()
    if p.returncode != 0:
        raise PdError('pdascii exited with error:\n[stderr]\n{0:s}\n'
                      '[stdout]\n{0:s}'.format(stderr, stdout))
    if not path.isfile(outfn):
        raise PdError('pdascii did not produce an output file:\n[stderr]\n{0:s}\n'
                      '[stdout]\n{0:s}'.format(stderr, stdout))
    return relfn, outfn


def render_pd_html(self, node, code, options, prefix='puredata', imgcls=None, alt=None):
    format = self.builder.config.pdgraph_format

    try:
        if format not in ('png', 'svg'):
            raise PdError("pdgraph_format must be one of 'png', 'svg', but is {0!r:s}".format(format))
        fname, outfn = render_pd(self, code, options, format, prefix)
    except PdError as exc:
        self.builder.warn('pd code %r: ' % code + str(exc))
        raise nodes.SkipNode

    if fname is None:
        self.body.append(self.encode(code))
    else:
        if alt is None:
            alt = node.get('alt', self.encode(code).strip())
        imgcss = imgcls and 'class="{0:s}"'.format(imgcls) or ''
        if format == 'svg':
            svgtag = '''<object data="{0:s}" type="image/svg+xml">
            <p class="warning">{1:s}</p></object>\n'''.format(fname, alt)
            self.body.append(svgtag)
        else:
            if 'align' in node:
                self.body.append('<div align="{0:s}" class="align-{1:s}">'
                                 .format(node['align'], node['align']))

            self.body.append('<img src="{0:s}" alt="{1:s}" {2:s}/>\n'.format(fname, alt, imgcss))
            if 'align' in node:
                self.body.append('</div>\n')

    raise nodes.SkipNode


def html_visit_pd(self, node):
    render_pd_html(self, node, node['code'], node['options'])


def render_pd_latex(self, node, code, options, prefix='puredata'):
    try:
        fname, outfn = render_pd(self, code, options, 'pdf', prefix)
    except PdError as exc:
        self.builder.warn('pd code %r: ' % code + str(exc))
        raise nodes.SkipNode

    is_inline = self.is_inline(node)
    if is_inline:
        para_separator = ''
    else:
        para_separator = '\n'

    if fname is not None:
        post = None
        if not is_inline and 'align' in node:
            if node['align'] == 'left':
                self.body.append('{')
                post = '\\hspace*{\\fill}}'
            elif node['align'] == 'right':
                self.body.append('{\\hspace*{\\fill}')
                post = '}'
        self.body.append('%s\\includegraphics{%s}%s' %
                         (para_separator, fname, para_separator))
        if post:
            self.body.append(post)

    raise nodes.SkipNode


def latex_visit_graphviz(self, node):
    render_pd_latex(self, node, node['code'], node['options'])


def setup(app):
    app.add_node(PdNode,
                 html=(html_visit_pd, None),
                 latex=(latex_visit_graphviz, None))
    app.add_directive('pd', PureData)
    app.add_config_value('pdgraph', 'pdascii', 'html')
    app.add_config_value('pdgraph_args', [], 'html')
    app.add_config_value('pdgraph_format', 'png', 'html')
    return {'version': sphinx.__display_version__, 'parallel_read_safe': False}



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2016 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

from __future__ import print_function
from ext_lexer import parse_string
import os
import re
import sys
import pprint
from mako.template import Template
from pddoc.txt import Parser
from pddoc.pd import Canvas, PdExporter, PdObject
from pddoc import CairoPainter
import argparse


def extract_doc_comment(data):
    res = list()
    is_doc = False

    for line in data:
        if re.match(r'\s*\/\*(\*|\!)', line):
            is_doc = True

        if re.match(r'\s*\*+\/', line):
            res.append(line)
            break

        if is_doc:
            res.append(line)

    return res


def extract_doc_example(data):
    res = list()
    is_example = False

    for line in data:
        if re.match(r'\s*\/\/\/\s*(@|\\)pddoc\s+', line):
            is_example = True

        if re.match(r'\s*\/\/\/\s*(@|\\)pddoc_end', line):
            res.append(line)
            break

        if is_example:
            res.append(line)

    return res


def graph_source(data):
    if len(data) < 2:
        return ''

    res = list()
    for line in data[1:-1]:
        res.append(re.sub(r'\s*\/{2,3}\s+', '', line))
    return ''.join(res)


def process_file(fname):
    if not os.path.exists(fname):
        print("file not exists: \"{0}\"".format(fname), file=sys.stderr)
        return None

    with open(fname, 'r') as myfile:
        res = {}
        data = myfile.readlines()

        doc = extract_doc_comment(data)
        if doc:
            res_doc = parse_string(''.join(doc))
            if res_doc:
                res['doc'] = res_doc

        example = extract_doc_example(data)
        if example:
            res['example'] = graph_source(example)

        return res


def generate_sphynx(data):
    main = data['doc'][0]
    info = data['doc'][1]
    res = '## {0}:{1}\n\n'.format(info['library'], info['name'])
    res += '{0}\n\n'.format(main)

    if 'args' in info:
        res += '### arguments\n\n'
        for i in info['args']:
            res += '# {0}\n'.format(i['args'])

    if 'inlet' in info:
        res += '### inlets\n\n'
        for i in info['inlet']:
            res += '# {0}\n'.format(i['info'])

    if 'outlet' in info:
        res += '\n### outlets\n\n'
        for i in info['outlet']:
            res += '# {0}\n'.format(i['info'])

    # res += '.. version: {0}\n'.format(data['version'])
    return res


def generate_pddoc(data, args):
    # template config
    tmpl_path = "{0:s}/../share/extract.pddoc".format(os.path.dirname(os.path.abspath(__file__)))
    pddoc_template = Template(filename=tmpl_path)
    doc = data['doc'][1]


    template_data = {}
    template_data['name'] = doc['name']
    template_data['description'] = doc.get('brief')
    template_data['license'] = doc.get('license', 'Unknown')
    template_data['library'] = doc.get('library', 'misc')
    template_data['version'] = doc.get('version', '0.0.0')
    template_data['category'] = doc.get('category', '')
    template_data['website'] = doc.get('website', '')
    template_data['authors'] = doc['author'].split(',')
    template_data['example'] = data.get('example')

    if 'inlet' in doc:
        PdObject.xlet_calculator.mem_db.add_object(doc['name'], [0], [0])

    ext_name = doc.get('name', 'unknown')
    xlet_db = "example_{0}.db".format(ext_name)

    # save xlet db
    with open(xlet_db, 'w') as db:
        str = doc['name'] + "\t"

        if doc['inlet']:
            for i in doc['inlet']:
                if i['type'] == 'control':
                    str += '.'
                elif i['type'] == 'sound':
                    str += '~'
                else:
                    print("unknown inlet type: {0}".format(i['type']))
        else:
            str += '-'

        str += '\t'

        if doc['outlet']:
            for i in doc['inlet']:
                if i['type'] == 'control':
                    str += '.'
                elif i['type'] == 'sound':
                    str += '~'
                else:
                    print("unknown outlet type: {0}".format(i['type']))
        else:
            str += '-'

        str += '\n'

        db.write(str)

    if 'see' in doc:
        template_data['see_also'] = doc['see'].split(',')

    # keywords = self._keywords,
    # css_file = self._css_file,
    # css = self._css,
    # aliases = self._aliases,
    # version = self._version,
    # inlets = self._inlets,
    # outlets = self._outlets,
    # arguments = self._arguments,
    # contacts = self._contacts,

    return pddoc_template.render(**template_data)


def main():
    arg_parser = argparse.ArgumentParser(description='PureData C/C++ comment extractor')
    arg_parser.add_argument('--xdens', metavar='xdens', default=1, type=float,
                            help="horizontal space between objects")
    arg_parser.add_argument('--ydens', metavar='ydens', default=1, type=float,
                            help="vertical space between objects")
    arg_parser.add_argument('input', metavar='SOURCE', help="C/C++ extension source file")
    args = vars(arg_parser.parse_args())

    res = process_file(args['input'])
    if not res:
        sys.exit(1)

    print(generate_pddoc(res, args))


if __name__ == '__main__':
    main()




#!/usr/bin/env python
# coding=utf-8
from __future__ import print_function
import argparse
import os
import logging
from pddoc.library import LibraryMaker

#   Copyright (C) 2016 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #


def check_file(fname):
    if os.path.exists(fname):
        return True
    else:
        logging.error("file not found: '%s'", fname)
        return False


def main():
    arg_parser = argparse.ArgumentParser(description='Generates Pd XML library documentation from input '
                                                     'pddoc files for individual objects.')
    arg_parser.add_argument('--library', '-l', metavar='NAME', required=True, help="library name")
    arg_parser.add_argument('--search-path', '-p', metavar='search_path', help="Adds search path for XInclude files")
    arg_parser.add_argument('--output', '-o', metavar='OUTPUT', help="output xml file")
    arg_parser.add_argument('--version', '-v', metavar='version', default="0.0", help="library version")
    arg_parser.add_argument('pddoc_files', metavar='FILE', nargs='+', help="pddoc files")

    args = vars(arg_parser.parse_args())
    files = filter(check_file, args['pddoc_files'])

    if not files:
        logging.error("no files exists...")
        exit(-1)

    lib = LibraryMaker(args['library'])
    lib.version = args['version']

    if args['search_path']:
        lib.add_search_path(args['search_path'])

    lib.process_files(files)
    lib.sort()

    xml_data = str(lib)

    if args['output']:
        with open(args['output'], 'w') as f:
            f.write(xml_data)
    else:
        print(xml_data)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

 
__author__ = 'Serge Poltavski'

import logging
import os.path as path
from ..colorformatter import ColorizingStreamHandler

root = logging.getLogger()
root.setLevel(logging.DEBUG)
root.addHandler(ColorizingStreamHandler())

EXTERNALS_DIR = path.join(path.dirname(__file__), "externals")
XLET_MESSAGE, XLET_SOUND, XLET_GUI, XLET_IGNORE = range(0, 4)

from obj import PdObject
from baseobject import BaseObject
from canvas import Canvas
from drawstyle import DrawStyle
from parser import Parser
from brectcalculator import BRectCalculator
from coregui import CoreGui
from message import Message
from comment import Comment
from factory import make_by_name
from xletcalculator import XletCalculator
from pdexporter import PdExporter
from array import Array



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2014 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

from obj import PdObject
from . import XLET_MESSAGE
from abstractvisitor import AbstractVisitor


class Message(PdObject):
    def __init__(self, x, y, atoms):
        super(Message, self).__init__("msg", x, y, 0, 0, atoms)

    def __str__(self):
        res = "[%-40s {x:%i,y:%i,id:%i}" % (self.args_to_string() + "(", self._x, self._y, self._id)
        return res

    def to_string(self):
        return self.args_to_string()

    def draw(self, painter):
        painter.draw_message(self)

    def inlets(self):
        return [XLET_MESSAGE]

    def outlets(self):
        return [XLET_MESSAGE]

    def traverse(self, visitor):
        assert isinstance(visitor, AbstractVisitor)

        if visitor.skip_message(self):
            return

        visitor.visit_message(self)



#!/usr/bin/env python
# coding=utf-8
import argparse
import os.path

from pddoc.cairopainter import *
from pddoc.pd import Parser
from format import detect_format

# Copyright (C) 2014 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #


__author__ = 'Serge Poltavski'


def main():
    arg_parser = argparse.ArgumentParser(description='Converts Pd patch to image')
    arg_parser.add_argument('--format', '-f', metavar='format', nargs=1, choices=("png", "pdf", "svg"),
                            help='output image format')
    arg_parser.add_argument('--width', metavar='px', nargs=1, help='image width in pixels')
    arg_parser.add_argument('--height', metavar='px', nargs=1, help='image height in pixels')
    arg_parser.add_argument('input', metavar='PATCH')
    arg_parser.add_argument('output', metavar='OUTNAME', nargs='?', default='image.png')

    args = vars(arg_parser.parse_args())

    pd_parser = Parser()

    finput = args['input']
    if not os.path.exists(finput):
        logging.error("File not exists: \"%s\"", finput)
        exit(1)

    pd_parser.parse(finput)

    width = pd_parser.canvas.width
    height = pd_parser.canvas.height

    if args['width'] is not None:
        width = int(args['width'][0])

    if args['height'] is not None:
        width = int(args['height'][0])

    fmt = detect_format(args)
    painter = CairoPainter(width, height, args['output'], fmt)
    pd_parser.canvas.draw(painter)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2016 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

from idocobjectvisitor import IDocObjectVisitor
from pd.obj import PdObject


class XletDocVisitor(IDocObjectVisitor):
    def __init__(self, write_to_file=True, add_to_mem_db=True):
        IDocObjectVisitor.__init__(self)
        self._inlet_types = []
        self._outlet_types = []
        self._write_to_file = write_to_file
        self._add_to_mem_db = add_to_mem_db
        self._dynamic_inlets = False
        self._dynamic_outlets = False
        self._aliases = []

    def inlets_begin(self, inlets):
        self._inlet_types = inlets.pd_type_list()
        self._dynamic_inlets = inlets.is_dynamic()

    def aliases_begin(self, a):
        self._aliases = a.aliases()

    def outlets_begin(self, outlets):
        self._outlet_types = outlets.pd_type_list()
        self._dynamic_outlets = outlets.is_dynamic()

    def inlet_types(self):
        if self._dynamic_inlets:
            return [3]  # XLET_IGNORE

        return self._inlet_types

    def outlet_types(self):
        if self._dynamic_outlets:
            return [3]  # XLET_IGNORE

        return self._outlet_types

    @staticmethod
    def as_db_type(xlets):
        if len(xlets) < 1:
            return '-'

        def x_type(x):
            return ('.', '~', '_', '?')[x]

        return ''.join(map(x_type, xlets))

    def names(self):
        return ",".join([self.name] + self._aliases)

    def object_end(self, obj):
        if self._write_to_file:
            fname = "{0}-xlet_db.txt".format(self.name)
            with open(fname, "w") as f:
                f.write("{0}\t\t{1}\t\t{2}\n".format(self.names(),
                                                     self.as_db_type(self.inlet_types()),
                                                     self.as_db_type(self.outlet_types())))

        if self._add_to_mem_db:
            PdObject.add_object_xlet_info(self.name, self.inlet_types(), self.outlet_types())



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

import os
import re
import sys
import logging

from floatatom import FloatAtom
from . import EXTERNALS_DIR
from obj import PdObject
from bng import PdBng
from toggle import PdToggle
from slider import PdSlider, PdHSlider, PdVSlider
from radio import Radio, PdVRadio, PdHRadio
from gcanvas import GCanvas
from nbx import Nbx
from vu import PdVu
from symbolatom import PdSymbolAtom

externals = {}
not_found = set()
imports = []


def _clean_ext_name(name):
    if '.' in name:
        name = name.replace('.', '_')
    if '~' in name:
        name = name.replace('~', '')

    return name


def make(atoms):
    assert isinstance(atoms, list)
    assert len(atoms) > 0
    name = atoms[0]

    if name == "floatatom":
        return FloatAtom.from_atoms(atoms[1:])
    elif name == "symbolatom":
        return PdSymbolAtom.from_atoms(atoms[1:])
    # elif name == "declare":
    #     print atoms
    #     assert False
    elif name == "bng":
        return PdBng.from_atoms(atoms[1:])
    elif name == "tgl":
        return PdToggle.from_atoms(atoms[1:])
    elif name in ("hsl", "vsl"):
        return PdSlider.from_atoms(atoms)
    elif name in ("hradio", "vradio"):
        return Radio.from_atoms(atoms)
    elif name == "cnv":
        return GCanvas.from_atoms(atoms[1:])
    elif name == "vu":
        return PdVu.from_atoms(atoms)
    elif name == "nbx":
        return Nbx.from_atoms(atoms)
    elif name not in not_found:
        if find_external_object(name):
            return externals[_clean_ext_name(name)].create(atoms)
    else:
        pass

    # handle import
    if name == "import":
        add_import(atoms[1])

    return PdObject(name, 0, 0, 0, 0, atoms[1:])


def add_import(name):
    import_path = os.path.join(EXTERNALS_DIR, name)
    if import_path in imports:
        return

    if os.path.exists(import_path):
        logging.debug("import path added: \"%s\"", import_path)
        imports.append(import_path)
    else:
        logging.warning("import path not found: \"%s\"", name)


def make_by_name(name, args=None, **kwargs):
    if args is None:
        args = []
    if name == "floatatom":
        return FloatAtom(0, 0, **kwargs)
    elif name == "symbolatom":
        return PdSymbolAtom(0, 0, **kwargs)
    elif name == "bng":
        return PdBng(0, 0, **kwargs)
    elif name == "tgl":
        return PdToggle(0, 0, **kwargs)
    elif name == "nbx":
        return Nbx(0, 0, **kwargs)
    elif name == "hslider":
        return PdHSlider(0, 0, **kwargs)
    elif name == "vslider":
        return PdVSlider(0, 0, **kwargs)
    elif name == "hradio":
        return PdHRadio(0, 0, **kwargs)
    elif name == "vradio":
        return PdVRadio(0, 0, **kwargs)
    elif name == "vu":
        return PdVu(0, 0, **kwargs)
    elif name not in not_found and find_external_object(name):
        return externals[_clean_ext_name(name)].create_by_name(name, args, **kwargs)
    else:
        return PdObject(name, 0, 0, 0, 0, args)


def _find_in_externals(name):
    mod_path = os.path.join(EXTERNALS_DIR, name)
    if os.path.exists(mod_path + ".py"):
        return mod_path
    else:
        return None


def _find_in_imports(name):
    for path in imports:
        mod_path = os.path.join(path, name)
        if os.path.exists(mod_path + ".py"):
            return mod_path

    return None


def find_external_object(name):
    clean_ext_name = _clean_ext_name(name)

    if clean_ext_name in externals:
        return True

    rname = re.compile(r"^([-a-zA-Z0-9~/*=+><!_%|&.]+)$")
    if not rname.match(clean_ext_name):
        logging.warning("name contains invalid characters: [%s]", clean_ext_name)
        return False

    mod_path = _find_in_externals(clean_ext_name)
    if not mod_path:
        mod_path = _find_in_imports(clean_ext_name)
    if not mod_path:
        not_found.add(clean_ext_name)
        return False

    mod_dir = os.path.dirname(mod_path)
    mod_name = os.path.basename(mod_path)

    if mod_dir not in sys.path:
        sys.path.append(mod_dir)
    try:
        mod = __import__(mod_name)
        externals[clean_ext_name] = mod
        logging.debug("module \"%s.py\" imported from \"%s\".", mod_name, mod_dir)
        return True
    except ImportError as e:
        logging.error("error while importing extension: %s - \"%s\". Paths: = %s", mod_name, e, "\n\t".join(sys.path))
        not_found.add(clean_ext_name)
        return None



#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #


__author__ = 'Serge Poltavski'

from coregui import CoreGui
from . import XLET_GUI

# [width]? - horizontal size of element
# [height]? - vertical size of element
# [receive]? - receive symbol name
# [label]? - label
# [x_off]? - horizontal position of the label text relative to the upperleft corner of the object
# [y_off]? - vertical position of the label text relative to the upperleft corner of the object
# [font]? - font type
# [fontsize]? - font size
# [bg_color]? - background color
# [label_color]? - label color
# [scale]? - when set the logarithmic scale is displayed
# [?]? - unknown value, default is zero


class PdVu(CoreGui):
    def __init__(self, x, y, **kwargs):
        CoreGui.__init__(self, "vu", x, y, [], **kwargs)
        self._width = int(kwargs.get("width", 15))  # number of digits the element displays
        self._height = int(kwargs.get("height", 120))  # vertical size of element in pixels
        self._scale = bool(int(kwargs.get("scale", 0)))  # when set the logarithmic scale is displayed

    @staticmethod
    def from_atoms(atoms):
        assert len(atoms) == 13
        assert atoms[0] == "vu"

        return PdVu(0, 0, width=atoms[1],
                    height=atoms[2],
                    receive=atoms[3],
                    label=atoms[4],
                    label_xoff=atoms[5],
                    label_yoff=atoms[6],
                    font_type=atoms[7],
                    font_size=atoms[8],
                    bg_color=atoms[9],
                    label_color=atoms[10],
                    scale=atoms[11])

    def inlets(self):
        return [XLET_GUI] * 2

    def outlets(self):
        return [XLET_GUI] * 2

    def get_width(self):
        return self._width + 2

    def get_height(self):
        return self._height

    def draw_scale(self, painter):
        scale_txt = (">+12", "+6", "+2", "-0dB", "-2", "-6", "-12", "-20", "-30", "-50", "<-99")
        x = self.right + 5
        y = self.top + 4
        ystep = self.get_height() / (len(scale_txt) - 1)

        for txt in scale_txt:
            painter.draw_text(x, y, txt, color=self.lbcolor())
            y += ystep

    def draw(self, painter):
        self.draw_bbox(painter)
        self.draw_label(painter)
        self.draw_xlets(painter)

        if self._scale:
            self.draw_scale(painter)



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2014 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                            #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

import obj
from termcolor import colored
import six
import logging

from . import XLET_GUI
from abstractvisitor import AbstractVisitor


class Color:
    def __init__(self, r=0, g=0, b=0):
        assert isinstance(r, float) or isinstance(r, int)
        assert isinstance(g, float) or isinstance(g, int)
        assert isinstance(b, float) or isinstance(b, int)
        self._r = r
        self._g = g
        self._b = b

    @staticmethod
    def black():
        return Color(0, 0, 0)

    @staticmethod
    def white():
        return Color(255, 255, 255)

    def rgb(self):
        return self._r, self._g, self._b

    def set_rgb(self, rgb):
        self._r = rgb[0]
        self._g = rgb[1]
        self._b = rgb[2]

    def from_pd(self, value):
        v = abs(int(value) + 1)
        r = v / 4096
        g = (v % 4096) / 64
        b = (v % 4096) % 64
        self._r = int(round(r / 63.0 * 255)) % 256
        self._g = int(round(g / 63.0 * 255)) % 256
        self._b = int(round(b / 63.0 * 255)) % 256

    @staticmethod
    def from_hex(hex_str):
        c = Color()

        if hex_str[0] == '#':
            hex_str = hex_str[1:]

        if len(hex_str) == 6:
            r = int(hex_str[0:2], 16)
            g = int(hex_str[2:4], 16)
            b = int(hex_str[4:6], 16)
            c.set_rgb((r, g, b))
        elif len(hex_str) == 3:
            r = int(hex_str[0], 16) * 0x11
            g = int(hex_str[1], 16) * 0x11
            b = int(hex_str[2], 16) * 0x11
            c.set_rgb((r, g, b))
        else:
            assert False

        return c

    def compare(self, rgb):
        if isinstance(rgb, list) or isinstance(rgb, tuple):
            return self._r == rgb[0] and self._g == rgb[1] and self._b == rgb[2]
        elif isinstance(rgb, self.__class__):
            return self._r == rgb._r and self._g == rgb._g and self._b == rgb._b
        else:
            return False

    def is_black(self):
        return self.compare((0, 0, 0))

    def rgb_float(self):
        return round(self._r / 255.0, 5), round(self._g / 255.0, 5), round(self._b / 255.0, 5)

    def to_pd(self):
        return -4096 * int(round(self._r * (63 / 255.0))) \
               - 64 * int(round(self._g * (63 / 255.0))) \
               - 1 * int(round(self._b * (63 / 255.0))) - 1

    def to_hex_str(self):
        return "#{0:02X}{1:02X}{2:02X}".format(self._r, self._g, self._b)

    def __str__(self):
        res = colored("RGB[%i,%i,%i]" % (self._r, self._g, self._b), "green")
        return res

    def __eq__(self, other):
        return self.compare(other)


def color_from_str(value):
    c = Color()

    if isinstance(value, int):
        c.from_pd(value)
    elif isinstance(value, six.string_types):
        if value[0] == '#':
            return Color.from_hex(value)
        else:
            c.from_pd(value)
    else:
        assert False

    return c


class CoreGui(obj.PdObject):
    POS_LEFT, POS_RIGHT, POS_TOP, POS_BOTTOM = (0, 1, 2, 3)

    @classmethod
    def is_coregui(cls, name):
        return name in ("bng", "tgl", "floatatom", "symbolatom",
                        "nbx", "hslider", "vslider",  "vsl", "hsl", "vu",
                        "hrd", "vrd", "hradio", "vradio")

    def __init__(self, name, x, y, args, **kwargs):
        obj.PdObject.__init__(self, name, x, y, 0, 0, args)
        self._send = kwargs.get("send", "empty")
        self._receive = kwargs.get("receive", "empty")
        self._label = kwargs.get("label", "empty")
        self._label_xoff = int(kwargs.get("label_xoff", 17))
        self._label_yoff = int(kwargs.get("label_yoff", 7))
        self._font_type = int(kwargs.get("font_type", 0))
        self._font_size = int(kwargs.get("font_size", 10))
        self._bg_color = color_from_str(kwargs.get("bg_color", -262144))
        self._fg_color = color_from_str(kwargs.get("fg_color", -1))
        self._label_color = color_from_str(kwargs.get("label_color", -1))

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, v):
        self._label = v

    @property
    def send(self):
        return self._send

    @send.setter
    def send(self, v):
        self._send = v

    def no_label(self):
        return not self._label or self.label == "empty"

    @property
    def receive(self):
        return self._receive

    @receive.setter
    def receive(self, v):
        self._receive = v

    def __str__(self):
        return "[GUI:%-36s {x:%i,y:%i,id:%i}" % (self.name + "]", self._x, self._y, self._id)

    def bgcolor(self):
        return self._bg_color.rgb_float()

    def fgcolor(self):
        return self._fg_color.rgb_float()

    def lbcolor(self):
        return self._label_color.rgb_float()

    def inlets(self):
        return [XLET_GUI]

    def outlets(self):
        return [XLET_GUI]

    def traverse(self, visitor):
        assert isinstance(visitor, AbstractVisitor)

        if visitor.skip_core_gui(self):
            return

        visitor.visit_core_gui(self)

    def draw_bbox(self, painter, **kwargs):
        vertexes = (
            (self.x, self.y),
            (0, self.height),
            (self.width, 0),
            (0, -self.height)
        )

        painter.draw_poly(vertexes, fill=self.bgcolor(), outline=(0, 0, 0), **kwargs)

    def show_inlets(self):
        return not self.receive or self.receive == "empty"

    def show_outlets(self):
        return not self.send or self.send == "empty"

    def draw_xlets(self, painter):
        if self.show_inlets():
            painter.draw_inlets(self.inlets(), self.x, self.y, self.width)
        if self.show_outlets():
            painter.draw_outlets(self.outlets(), self.x, self.bottom, self.width)

    def draw_label(self, painter):
        if not self.no_label():
            lx, ly = self._get_label_xy()
            ly += self._font_size / 2  # TODO hardcoded
            painter.draw_text(lx, ly, self.label,
                              color=self.lbcolor(),
                              font_size=self._font_size,  # FIXME! hardcoded font family
                              font="Menlo")

    def draw_on_parent(self):
        return True

    def _get_label_xy(self):
        return self.x + self._label_xoff, self.y + self._label_yoff  # font height correction

    def _get_label_pos(self, pos):
        if isinstance(pos, int):
            if pos not in (self.POS_LEFT, self.POS_TOP, self.POS_BOTTOM, self.POS_RIGHT):
                logging.warning("invalid label position: {0:d}".format(pos))
                return None

            return pos
        elif isinstance(pos, six.string_types):
            lmap = {"left": self.POS_LEFT,
                    "right": self.POS_RIGHT,
                    "top": self.POS_TOP,
                    "bottom": self.POS_BOTTOM}

            if pos.lower() in lmap:
                return lmap[pos.lower()]
            else:
                logging.warning("invalid label position: {0:s}".format(pos))
                return None
        else:
            logging.warning("invalid label position: {0:s}".format(str(pos)))
            return None

    def to_atoms(self):
        return {}



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2016 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

import argparse
import logging
import os

from pddoc.txt import Parser
from pddoc.pd import Canvas, PdExporter, PdObject
from pddoc import CairoPainter
import pddoc.pd.factory as factory


def main():
    fmt_list = ("png", "pdf", "svg", "pd")
    arg_parser = argparse.ArgumentParser(description='PureData ascii doc to Pd/PDF/SVG converter')
    arg_parser.add_argument('--auto', '-a', help='calculate output image size', action='store_true')
    arg_parser.add_argument('--format', '-f', metavar='FMT', nargs=1, choices=fmt_list,
                            help='output format ({0})'.format(",".join(fmt_list)), default=["pd"])
    arg_parser.add_argument('--width', '-wd', metavar='X', type=int, nargs=1,
                            help='image width in pixels', default=400)
    arg_parser.add_argument('--height', '-ht', metavar='X', type=int, nargs=1,
                            help='image height in pixels', default=300)
    arg_parser.add_argument('--xlet-db', metavar='PATH', action='append',
                            help='inlet/outlet database file path', default=[])
    arg_parser.add_argument('input', metavar='INPUT', help="Documentation file in pd ascii format")
    arg_parser.add_argument('output', metavar='OUTPUT', nargs='?', default='',
                            help="output file name")

    args = vars(arg_parser.parse_args())
    in_file = args['input']
    output = args['output']
    wd = args['width']
    ht = args['height']
    fmt = args['format'][0]

    if not os.path.exists(in_file):
        logging.error("File not exists: \"%s\"", in_file)
        exit(1)

    for db_path in args['xlet_db']:
        if not os.path.exists(db_path):
            logging.warning("xlet database file not found: '%s'. skipping...", db_path)
        else:
            PdObject.xlet_calculator.add_db(db_path)

    if not output:
        output = os.path.splitext(os.path.basename(in_file))[0] + "." + fmt

    factory.add_import('ceammc')

    p = Parser()
    p.parse_file(in_file)

    cnv = Canvas(0, 0, wd, ht)
    cnv.type = Canvas.TYPE_WINDOW
    p.export(cnv)

    if fmt in ('png', 'pdf', 'svg'):
        if args['auto']:
            br_calc = cnv.brect_calc()
            cnv.traverse(br_calc)
            bbox = br_calc.brect()
            wd = bbox[2] + Parser.X_PAD * 2
            ht = bbox[3] + Parser.Y_PAD * 2

        painter = CairoPainter(wd, ht, output, fmt)
        cnv.draw(painter)
    elif fmt == 'pd':
        pd_exporter = PdExporter()
        cnv.traverse(pd_exporter)
        pd_exporter.save(output)
    else:
        print("Unknown output format: {0:s}".format(fmt))

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2015 by Serge Poltavski                                 #
# serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #


__author__ = 'Serge Poltavski'

from coregui import *


class Nbx(CoreGui):
    def __init__(self, x, y, **kwargs):
        CoreGui.__init__(self, "nbx", x, y, [], **kwargs)
        self._digits = int(kwargs.get("digits", 5))  # number of digits the element displays
        self._height = int(kwargs.get("height", 14))  # vertical size of element in pixels
        self._min = float(kwargs.get("min", -1e+037))  # minimum value
        self._max = float(kwargs.get("max", 1e+037))  # maximum value
        self._log = int(kwargs.get("log", 0))  # linear when unset, logarithmic when set
        self._init = float(kwargs.get("init", 0))
        # logarithmic steps, accepts values from 10 to 2000, default is 256
        self._log_height = int(kwargs.get("log_height", 256))

    @staticmethod
    def from_atoms(atoms):
        assert len(atoms) == 19
        assert atoms[0] == "nbx"

        return Nbx(0, 0, digits=atoms[1],
                   height=atoms[2],
                   min=atoms[3],
                   max=atoms[4],
                   log=atoms[5],
                   init=atoms[6],
                   send=atoms[7],
                   receive=atoms[8],
                   label=atoms[9],
                   label_xoff=atoms[10],
                   label_yoff=atoms[11],
                   font_type=atoms[12],
                   font_size=atoms[13],
                   bg_color=atoms[14],
                   fg_color=atoms[15],
                   label_color=atoms[16],
                   log_height=atoms[18])

    def get_width(self):
        return (self._digits + 1) * self._font_size * 0.85

    def get_height(self):
        return self._height

    def draw_value(self, painter):
        h = self.height / 2
        painter.draw_line(self.left, self.top, self.left + h, self.top + h, color=self.fgcolor())
        painter.draw_line(self.left, self.bottom, self.left + h, self.top + h, color=self.fgcolor())

        yoff = self._font_size / 5
        painter.draw_text(self.left + h + 2, self.bottom - yoff, str(self._init),
                          color=self.fgcolor(),
                          font="Menlo",
                          font_size=self._font_size)

    def draw(self, painter):
        skos = 4
        vertexes = (
            (self.x, self.y),
            (0, self.height),
            (self.width, 0),
            (0, -self.height + skos),
            (-skos, -skos)
        )

        painter.draw_poly(vertexes, fill=self.bgcolor(), outline=(0, 0, 0))

        self.draw_value(painter)
        self.draw_label(painter)
        self.draw_xlets(painter)



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2016 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

from pd.canvas import Canvas
from pd.gcanvas import GCanvas
from pd.coregui import Color
from pd.brectcalculator import BRectCalculator
from pd.externals.pddp.pddplink import PddpLink
from pd.externals.ceammc.ui_link import UILink
from pd.comment import Comment
from pd.pdexporter import PdExporter
import re


class PdPageStyle(object):
    HRULE_COLOR = Color(200, 200, 200)
    SECTION_FONT_SIZE = 17
    SECTION_FONT_COLOR = Color(50, 50, 50)
    HEADER_HEIGHT = 40
    HEADER_FONT_SIZE = 20
    HEADER_BG_COLOR = Color(100, 100, 100)
    HEADER_FONT_COLOR = Color(0, 255, 255)
    FOOTER_HEIGHT = 40
    FOOTER_BG_COLOR = Color(200, 200, 200)
    INFO_BG_COLOR = Color(240, 255, 255)


class PdPage(object):
    brect_calc = BRectCalculator()

    def __init__(self, title, width=600, height=500):
        self._title = title.replace(' ', '_')
        self._width = int(width)
        self._height = int(height)
        self._canvas = Canvas(0, 0, self._width, self._height, font_size=12)
        self._canvas.type = Canvas.TYPE_WINDOW
        self._pd = {} # subpatches dict

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def canvas(self):
        return self._canvas

    def make_background(self, x, y, width, height, color):
        cnv = GCanvas(x, y, width=width, height=height, size=1)
        cnv._bg_color = color
        return cnv

    def make_hrule(self, x, y, width, height=1, color=Color(200, 200, 200)):
        return self.make_background(x, y, width, height, color)

    def make_styled_hrule(self, y):
        return self.make_hrule(20, y, width=self._width - 30, color=PdPageStyle.HRULE_COLOR)

    def make_label(self, x, y, txt, font_size, color=Color.black(), bg_color=Color.white()):
        w, h = self.brect_calc.string_brect(txt, font_size)[2:]

        cnv = GCanvas(x, y, width=w + 10, height=h * 1.6, size=5, font_size=font_size, label_xoff=4,
                      label_yoff=font_size - 4)
        cnv.label = txt.replace(' ', '_')
        cnv._label_color = color
        cnv._bg_color = bg_color
        return cnv

    def make_header(self, title):
        cnv = GCanvas(1, 1,
                      width=self._width - 3,
                      height=PdPageStyle.HEADER_HEIGHT,
                      size=5,
                      font_size=PdPageStyle.HEADER_FONT_SIZE,
                      label_xoff=PdPageStyle.HEADER_FONT_SIZE,
                      label_yoff=PdPageStyle.HEADER_FONT_SIZE)
        cnv.label = title.replace(' ', '_')
        cnv._label_color = PdPageStyle.HEADER_FONT_COLOR
        cnv._bg_color = PdPageStyle.HEADER_BG_COLOR
        return cnv

    def make_footer(self, bottom, height=PdPageStyle.FOOTER_HEIGHT):
        y = max(bottom, self._height - height - 2)
        cnv = GCanvas(1, y,
                      width=self._width - 3,
                      height=height,
                      size=5)

        cnv._bg_color = PdPageStyle.FOOTER_BG_COLOR
        return cnv

    def make_section_label(self, y, txt, font_size=PdPageStyle.SECTION_FONT_SIZE):
        return self.make_label(20, y, txt, font_size, PdPageStyle.SECTION_FONT_COLOR)

    def make_section(self, y, txt):
        l = self.make_section_label(y, txt)
        r = self.make_styled_hrule(y + l.height + 10)
        return l, r, self.group_brect([l, r])

    def make_link(self, x, y, url, name):
        kw = {'@url': url, '@title': name}
        return UILink(x, y + 3, **kw)

    def make_txt(self, txt, x, y):
        if txt:
            txt = re.sub(' +', ' ', txt)
            #  match only number with end dot: 1. - used for enums
            txt = re.sub('(\d+)\\.(?!\d)', '\\1\\.', txt)
            txt = re.sub(' *, *', ' \\, ', txt)
        else:
            txt = ""

        return Comment(x, y, txt.split(' '))

    def make_subpatch(self, name, x, y, w, h):
        pd = Canvas(x, y, w, h, name=name)
        pd.type = Canvas.TYPE_SUBPATCH
        self._pd[name] = pd
        return pd

    def add_header(self, title):
        h = self.make_header(title)
        h.calc_brect()
        self._canvas.append_object(h)
        return h

    def add_to_subpatch(self, name, obj):
        assert name in self._pd
        self._pd[name].append_object(obj)

    def add_pd_txt(self, pd, txt, x, y):
        t = self.make_txt(txt, x, y)
        pd.append_object(t)
        t.calc_brect()
        return t

    def add_txt(self, txt, x, y):
        return self.add_pd_txt(self._canvas, txt, x, y)

    def add_link(self, txt, url, x, y):
        lnk = self.make_link(x, y, url, txt)
        self._canvas.append_object(lnk)
        return lnk

    def add_bg_txt(self, txt, color, x, y):
        t = self.make_txt(txt, x, y)
        t.calc_brect()
        bg = self.make_background(x, y, t.width + 8, t.height + 8, color=color)
        self._canvas.append_object(bg)
        self._canvas.append_object(t)
        return t, bg

    def add_description(self, txt, y):
        d = self.add_bg_txt(txt, PdPageStyle.INFO_BG_COLOR, 0, y)
        br = self.group_brect(d)
        self.move_to_x(d, self._width - br[2] - 20)

    def add_section(self, title, y):
        l, r, brect = self.make_section(y, title)
        self._canvas.append_object(l)
        self._canvas.append_object(r)
        return l, r, brect

    def add_subpatch_txt(self, name, txt, x, y):
        return self.add_pd_txt(self._pd[name], txt, x, y)

    def add_subpatch_obj(self, name, obj):
        self._pd[name].append_object(obj)
        obj.calc_brect()
        return obj

    def append_object(self, obj):
        self._canvas.append_object(obj)
        obj.calc_brect()
        return obj

    def append_list(self, obj):
        for o in obj:
            self.append_object(o)

    def group_brect(self, seq):
        for el in seq:
            el.calc_brect()

        left = min(seq, key=lambda x: x.left).left
        top = min(seq, key=lambda x: x.top).top
        right = max(seq, key=lambda x: x.right).right
        bottom = max(seq, key=lambda x: x.bottom).bottom
        return left, top, right - left, bottom - top

    def move_to_x(self, seq, x):
        if len(seq) < 1:
            return

        x_off = x - seq[0].x

        for o in seq:
            o.x += x_off

    def move_to_y(self, seq, y):
        if len(seq) < 1:
            return

        y_off = y - seq[0].y
        for o in seq:
            o.y += y_off

    def move_to(self, seq, x, y):
        for o in seq:
            o.x = x
            o.y = y

    def move_by(self, seq, x, y):
        for o in seq:
            o.x += x
            o.y += y

    def move_by_x(self, seq, x):
        for o in seq:
            o.x += x

    def move_by_y(self, seq, y):
        for o in seq:
            o.x += y

    def bottom(self):
        for o in self._canvas.objects:
            o.calc_brect()

        max(lambda x: x.bottom, self._canvas.objects)

    def place_under(self, old_obj, new_obj, y_off=0):
        old_obj.calc_brect()
        new_obj.y = old_obj.bottom + y_off

    def place_right_side(self, old_obj, new_obj, x_space=0):
        old_obj.calc_brect()
        new_obj.x = old_obj.right + x_space

    def place_top_right(self, obj, x_pad=0, y_pad=0):
        obj.calc_brect()
        obj.x = self._width - obj.width - x_pad
        obj.y = y_pad

    def place_bottom_right(self, obj, x_pad=0, y_pad=0):
        obj.calc_brect()
        obj.x = self._width - obj.width - x_pad
        obj.y = self._height - obj.height - y_pad

    def place_bottom_left(self, obj, x_pad=0, y_pad=0):
        obj.calc_brect()
        obj.x = x_pad
        obj.y = self._height - obj.height - y_pad

    def place_in_row(self, seq, x, x_pad=0):
        for o in seq:
            o.calc_brect()

        for o in seq:
            o.x = x
            x += o.width + x_pad

    def place_in_col(self, seq, y, y_pad=0):
        for o in seq:
            o.calc_brect()

        for o in seq:
            o.y = y
            y += o.height + y_pad

    def to_string(self):
        pd_exporter = PdExporter()
        self._canvas.traverse(pd_exporter)
        return '\n'.join(pd_exporter.result[:-1])



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'


class IDocObjectVisitor(object):
    def __init__(self):
        self._name = ""

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, n):
        self._name = n

    def a_begin(self, a):
        pass

    def a_end(self, a):
        pass

    def aliases_begin(self, aliases):
        pass

    def aliases_end(self, aliases):
        pass

    def also_begin(self, also):
        pass

    def also_end(self, also):
        pass

    def argument_begin(self, arg):
        pass

    def argument_end(self, arg):
        pass

    def arguments_begin(self, args):
        pass

    def arguments_end(self, args):
        pass

    def author_begin(self, author):
        pass

    def author_end(self, author):
        pass

    def authors_begin(self, authors):
        pass

    def authors_end(self, authors):
        pass

    def category_begin(self, cat):
        pass

    def category_end(self, cat):
        pass

    def col_begin(self, col):
        pass

    def col_end(self, col):
        pass

    def contacts_begin(self, contacts):
        pass

    def contacts_end(self, contacts):
        pass

    def description_begin(self, descr):
        pass

    def description_end(self, descr):
        pass

    def example_begin(self, ex):
        pass

    def example_end(self, ex):
        pass

    def info_begin(self, info):
        pass

    def info_end(self, info):
        pass

    def inlet_begin(self, inlet):
        pass

    def inlet_end(self, inlet):
        pass

    def inlets_begin(self, inlets):
        pass

    def inlets_end(self, inlets):
        pass

    def itemize_begin(self, itemize):
        pass

    def itemize_end(self, itemize):
        pass

    def item_begin(self, item):
        pass

    def item_end(self, item):
        pass

    def keywords_begin(self, kw):
        pass

    def keywords_end(self, kw):
        pass

    def library_begin(self, lib):
        pass

    def library_end(self, lib):
        pass

    def license_begin(self, license):
        pass

    def license_end(self, license):
        pass

    def meta_begin(self, meta):
        pass

    def meta_end(self, meta):
        pass

    def object_begin(self, obj):
        pass

    def object_end(self, obj):
        pass

    def outlet_begin(self, outlet):
        pass

    def outlet_end(self, outlet):
        pass

    def outlets_begin(self, outlets):
        pass

    def outlets_end(self, outlets):
        pass

    def pdascii_begin(self, pdascii):
        pass

    def pdascii_end(self, pdascii):
        pass

    def pdcomment_begin(self, pdcomment):
        pass

    def pdcomment_end(self, pdcomment):
        pass

    def pdconnect_begin(self, pdconn):
        pass

    def pdconnect_end(self, pdconn):
        pass

    def pdexample_begin(self, pdexample):
        pass

    def pdexample_end(self, pdexample):
        pass

    def pdinclude_begin(self, pdinc):
        pass

    def pdinclude_end(self, pdinc):
        pass

    def pdmessage_begin(self, pdmsg):
        pass

    def pdmessage_end(self, pdmsg):
        pass

    def pdobject_begin(self, pdobj):
        pass

    def pdobject_end(self, pdobj):
        pass

    def row_begin(self, row):
        pass

    def row_end(self, row):
        pass

    def see_begin(self, see):
        pass

    def see_end(self, see):
        pass

    def title_begin(self, t):
        pass

    def title_end(self, t):
        pass

    def version_begin(self, v):
        pass

    def version_end(self, v):
        pass

    def website_begin(self, ws):
        pass

    def website_end(self, ws):
        pass

    def xinfo_begin(self, ws):
        pass

    def xinfo_end(self, ws):
        pass

    def since_begin(self, s):
        pass

    def since_end(self, s):
        pass

    def methods_begin(self, m):
        pass

    def methods_end(self, m):
        pass

    def method_begin(self, m):
        pass

    def method_end(self, m):
        pass

    def param_begin(self, p):
        pass

    def param_end(self, p):
        pass

    def par_begin(self, par):
        pass

    def par_end(self, par):
        pass

    def render(self):
        raise NotImplementedError

    def properties_begin(self, p):
        pass

    def properties_end(self, p):
        pass

    def property_begin(self, p):
        pass

    def property_end(self, p):
        pass



# coding=utf-8

import logging
from colorformatter import ColorizingStreamHandler
from .cairopainter import CairoPainter

#root = logging.getLogger()
#root.setLevel(logging.DEBUG)
#root.addHandler(ColorizingStreamHandler())



#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2016 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

from __future__ import print_function
import ply.lex as lex
from ply.lex import TOKEN
import re
import logging

# token names
tokens = (
    'OBJECT',
    'MESSAGE',
    'COMMENT',
    'CONNECTION',
    'CONNECTION_RIGHT',
    'CONNECTION_LEFT',
    'CONNECTION_X'
)

# token regexp
r_OBJECT = r'\[((?:[^\[\\]|\\.)+)\]'
r_MESSAGE = r'\[((?:[^\(\\]|\\.)+)\('
r_COMMENT = r'/\*(.+)\*/'
r_NEWLINE = r'\n+'
t_CONNECTION = r'\^*\|\.*'          # ^^|.. connection
t_CONNECTION_RIGHT = r'\^*\\_*\.*'  # ^^\______..
t_CONNECTION_LEFT = r'\.*_*/\^*'    # ^^\______..
t_CONNECTION_X = r'(x|X)'
t_ignore = ' \r\t\f'


def t_error(t):
    logging.error("Illegal character '{0:s}'".format(t.value[0]))
    t.lexer.skip(1)


@TOKEN(r_NEWLINE)
def t_newline(t):
    t.lexer.lineno += len(t.value)


@TOKEN(r_OBJECT)
def t_OBJECT(t):
    return t


@TOKEN(r_MESSAGE)
def t_MESSAGE(t):
    return t


@TOKEN(r_COMMENT)
def t_COMMENT(t):
    return t


def lexer():
    return lex.lex(reflags=re.UNICODE)



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

 
__author__ = 'Serge Poltavski'


from coregui import *


class PdSlider(CoreGui):
    def __init__(self, name, x, y, **kwargs):
        CoreGui.__init__(self, name, x, y, [], **kwargs)
        self._min = float(kwargs.get("min", 0))
        self._max = float(kwargs.get("max", 127))
        assert self._min <= self._max
        self._log = int(kwargs.get("log", 0))
        self._init = int(kwargs.get("init", 0))
        self._default_value = int(kwargs.get("default_value", 0))
        self._steady = int(kwargs.get("steady", 0))
        self._value = float(kwargs.get("value", self._min))
        self._pad = 3

    def slider_width(self):
        if self._value == self._max / 2:
            return 5
        else:
            return 3

    def slider_pos(self):
        assert self._value <= self._max
        return (self._value - self._min) / (self._max - self._min)

    @staticmethod
    def from_atoms(atoms):
        assert isinstance(atoms, list) and len(atoms) == 19
        assert atoms[0] in ("hsl", "hslider", "vsl", "vslider")

        if atoms[0] == "hslider":
            return PdHSlider(0, 0,
                             width=atoms[1],
                             height=atoms[2],
                             min=atoms[3],
                             max=atoms[4],
                             log=atoms[5],
                             init=atoms[6],
                             send=atoms[7],
                             receive=atoms[8],
                             label=atoms[9],
                             label_xoff=atoms[10],
                             label_yoff=atoms[11],
                             font_type=atoms[12],
                             font_size=atoms[13],
                             bg_color=atoms[14],
                             fg_color=atoms[15],
                             label_color=atoms[16],
                             default_value=atoms[17],
                             steady=atoms[18])
        else:
            return PdVSlider(0, 0,
                             width=atoms[1],
                             height=atoms[2],
                             min=atoms[3],
                             max=atoms[4],
                             log=atoms[5],
                             init=atoms[6],
                             send=atoms[7],
                             receive=atoms[8],
                             label=atoms[9],
                             label_xoff=atoms[10],
                             label_yoff=atoms[11],
                             font_type=atoms[12],
                             font_size=atoms[13],
                             bg_color=atoms[14],
                             fg_color=atoms[15],
                             label_color=atoms[16],
                             default_value=atoms[17],
                             steady=atoms[18])

    def to_atoms(self):
        return ["#X", "obj", self.x, self.y, self.name, self.width, self.height,
                self._min, self._max, self._log, self._init,
                self.send, self.receive,
                self.label, self._label_xoff, self._label_yoff,
                self._font_type, self._font_size,
                self._bg_color.to_pd(), self._fg_color.to_pd(), self._label_color.to_pd(),
                self._default_value, self._steady]


class PdHSlider(PdSlider):
    def __init__(self, x, y, **kwargs):
        PdSlider.__init__(self, "hslider", x, y, **kwargs)
        self.width = int(kwargs.get("width", 128))
        self.height = int(kwargs.get("height", 15))

    def draw(self, painter):
        self.draw_bbox(painter)
        self.draw_label(painter)

        xoff = self.slider_pos()
        x = int(self.left + self._pad + (self.width - self._pad) * xoff)
        painter.draw_line(x, self.top, x, self.bottom, color=self.fgcolor(), width=self.slider_width())

        self.draw_xlets(painter)


class PdVSlider(PdSlider):
    def __init__(self, x, y, **kwargs):
        PdSlider.__init__(self, "vslider", x, y, **kwargs)

        self.width = int(kwargs.get("width", 15))
        self.height = int(kwargs.get("height", 128))

    def draw(self, painter):
        self.draw_bbox(painter)
        self.draw_label(painter)

        yoff = self.slider_pos()
        y = int(self.bottom - self._pad - (self.height - self._pad) * yoff)
        painter.draw_line(self.left, y, self.right, y, color=self.fgcolor(), width=self.slider_width())

        self.draw_xlets(painter)



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

 
__author__ = 'Serge Poltavski'

from pddoc.pd import XLET_SOUND


_objects = {
    "crossfade~": (
        lambda args: ((1 if not args else int(args[0])) * 2 + 1) * [XLET_SOUND],
        lambda args: (1 if not args else int(args[0])) * [XLET_SOUND]
    )
}


def has_object(name):
    return name in _objects


def inlets(name, args):
    if name in _objects:
        return _objects[name][0](args)

    return []


def outlets(name, args):
    if name in _objects:
        return _objects[name][1](args)

    return []



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2016 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

from __future__ import print_function
from itertools import ifilter
import graph_lexer as lex
import re
from pddoc.pd import Message, Comment, Canvas, Array
from six import string_types
from pddoc.pd import factory
from pddoc.pd.coregui import CoreGui
import copy
import logging


class Node(object):
    def __init__(self, tok, char_pos):
        self.tok = tok
        self.char_pos_ = char_pos
        self.pd_object = None
        self.conn_src_id = None
        self.conn_src_outlet = 0
        self.conn_dest_id = None
        self.conn_dest_inlet = 0
        self.connected = False
        self.obj_line_index = -1

        if self.tok and self.is_connection():
            # calc connection source outlet
            self.conn_src_outlet = self.tok.value.count('^')
            self.conn_dest_inlet = self.tok.value.count('.')

            # only single \
            if self.tok.value == '\\':
                self.conn_dest_inlet += 1

    def group(self, n=1):
        return self.tok.lexmatch.group(n)

    def is_object(self):
        if self.tok is None:
            return False

        return self.tok.type in ('OBJECT', 'MESSAGE', 'COMMENT')

    def is_connection(self):
        return self.tok.type in ('CONNECTION', 'CONNECTION_LEFT', 'CONNECTION_RIGHT', 'CONNECTION_X', 'CONNECTION_MANUAL')

    @property
    def id(self):
        return self.line_pos * 1000 + self.char_pos

    @property
    def type(self):
        if self.tok is None:
            return None

        return self.tok.type

    @property
    def width(self):
        return len(self.tok.value)

    @property
    def value(self):
        return self.tok.value

    @property
    def line_pos(self):
        return self.tok.lineno - 1

    @property
    def char_pos(self):
        return self.char_pos_

    def contains(self, char_pos):
        return self.char_pos <= char_pos <= (self.char_pos + self.width)


class Parser(object):
    X_PAD = 20
    Y_PAD = 20
    X_SPACE = 8
    Y_SPACE = 12

    ALIASES = {
        '_': 'tgl',
        'T': 'tgl',
        'B': 'bng',
        'O': 'bng',
        'o': 'bng',
        'fa': 'floatatom',
        'F': 'floatatom',
        'S': 'symbolatom',
        'HR': 'hradio',
        'VR': 'vradio',
        'hsl': 'hslider',
        'vsl': 'vslider',
        'HS': 'hslider',
        'VS': 'vslider',
        'A': 'array'
    }

    def __init__(self):
        self.lines = []
        self.lines_len = []
        self.tokens = []
        self.nodes = []
        self.lexer = lex.lexer()

    def clear(self):
        self.lines = []
        self.lines_len = []
        self.tokens = []
        self.nodes = []
        self.lexer = lex.lexer()

    def parse_file(self, fname):
        with open(fname) as f:
            self.parse(f.read())

    def node_by_id(self, id):
        res = filter(lambda n: id == n.id, self.nodes)
        if len(res) < 1:
            return None

        return res[0]

    def parse(self, string):
        assert isinstance(string, string_types)

        self.lines = string.split('\n')
        self.lines_len = map(lambda x: len(x), self.lines)
        self.lexer.input(string)
        self.parse_tokens()
        self.parse_nodes()
        self.enumerate_objects()
        self.parse_connections()
        self.layout_nodes()
        return True

    def token_line_lex_pos(self, lexline, lexpos):
        assert lexpos >= 0
        assert lexline >= 0
        assert lexline < len(self.lines_len)

        return lexpos - sum(self.lines_len[0:lexline]) - lexline

    def enumerate_objects(self):
        for l in range(0, self.num_lines()):
            idx = 0
            for o in filter(lambda x: x.line_pos == l, self.nodes):
                o.obj_line_index = idx
                idx += 1

    def parse_tokens(self):
        while True:
            tok = self.lexer.token()
            if not tok:
                break

            self.tokens.append(tok)

            ln = tok.lineno - 1
            char_pos = self.token_line_lex_pos(ln, tok.lexpos)
            n = Node(tok, char_pos)
            self.nodes.append(n)

    @classmethod
    def find_alias(cls, atoms):
        name = atoms[0]
        args = atoms[1:]
        if name in cls.ALIASES:
            return cls.ALIASES[name], args
        return name, args

    def find_node_id_by_hash(self, hash):
        for n in filter(lambda x: x.is_object(), self.nodes):
            if n.value.startswith('X'):
                continue

            m = re.match(lex.r_OBJECT, n.value)
            if not m:
                m = re.match(lex.r_MESSAGE, n.value)
                if not m:
                    continue

            # filter spaces and #ID values
            atoms = filter(lambda a: len(a) > 0 and a.startswith('#'), m.group(1).split(' '))
            if '#' + hash in atoms:
                return n.id

        return False

    def parse_nodes(self):
        for n in filter(lambda x: x.is_object(), self.nodes):
            if n.type == 'OBJECT':
                m = re.match(lex.r_OBJECT, n.value)
                # filter spaces and #ID values
                atoms = filter(lambda a: len(a) > 0 and (not a.startswith('#')), m.group(1).split(' '))
                assert len(atoms) > 0
                name, args = self.find_alias(atoms)
                kwargs = dict()

                for a in args:
                    pair = a.split('=')
                    if len(pair) == 2:
                        kwargs[pair[0]] = pair[1]

                if CoreGui.is_coregui(name):
                    n.pd_object = factory.make_by_name(name, **kwargs)
                elif name == 'pd':
                    n.pd_object = Canvas.subpatch(args[0])
                elif name == 'X':
                    n.tok.type = 'CONNECTION_MANUAL'
                    str = atoms[1]
                    conn = []

                    for i in map(lambda x: x.split(':'), str.split('->')):
                        if len(i) == 1:
                            i.append(0)

                        assert len(i) == 2

                        obj_id = self.find_node_id_by_hash(i[0])
                        if obj_id is False:
                            logging.error("can't find node with id: {0:s}".format(i[0]))
                            return

                        i[0] = obj_id
                        conn.append(i)

                    n.conn_src_id = conn[0][0]
                    n.conn_src_outlet = int(conn[0][1])
                    n.conn_dest_id = conn[1][0]
                    n.conn_dest_inlet = int(conn[1][1])

                elif name == 'array':
                    n.pd_object = Array(args[0], kwargs.get('size', 100), kwargs.get('save', 0))
                    n.pd_object.width = kwargs.get('w', 200)
                    n.pd_object.height = kwargs.get('h', 140)
                    yr = map(lambda x: float(x), kwargs.get('yr', "-1..1").split('..'))
                    n.pd_object.set_yrange(yr[0], yr[1])
                else:
                    n.pd_object = factory.make_by_name(name, args, **kwargs)
            elif n.type == 'MESSAGE':
                m = re.match(lex.r_MESSAGE, n.value)
                txt = m.group(1).replace(',', '\,')
                args = filter(lambda a: len(a) > 0 and (not a.startswith('#')), txt.split(' '))
                n.pd_object = Message(0, 0, args)
            elif n.type == 'COMMENT':
                m = re.match(lex.r_COMMENT, n.value)
                txt = m.group(1).replace(';', ' \;').replace(',', ' \,')
                n.pd_object = Comment(0, 0, txt.split(' '))
            else:
                logging.warning("Unknown type {0:s}".format(n.type))

    def parse_connections(self):
        for c in filter(lambda n: n.is_connection(), self.nodes):
            self.process_connection(c)

    def find_connection(self, line, char_pos):
        """
        :type line: int
        :type char_pos: list
        """
        return next(ifilter(
            lambda x:
            x.line_pos == line and any(map(lambda c: x.contains(c), char_pos)), self.nodes), None)

    def find_by_line_idx(self, line, idx):
        return next((ifilter(lambda n: n.line_pos == line and n.obj_line_index == idx, self.nodes)), None)

    def process_cross_connection(self, conn):
        assert isinstance(conn, Node)
        line = conn.line_pos
        src_idx = conn.obj_line_index - 1
        dest_idx = conn.obj_line_index + 1

        src = self.find_by_line_idx(line, src_idx)
        dest = self.find_by_line_idx(line, dest_idx)

        if src and dest:
            c1 = copy.deepcopy(conn)
            c1.tok.type = 'CONNECTION'
            c1.conn_src_id = src.id
            c1.conn_dest_id = dest.id
            c1.conn_dest_inlet = 0
            c1.conn_src_outlet = len(src.pd_object.outlets()) - 1
            c1.connected = True

            c2 = copy.deepcopy(conn)
            c2.tok.type = 'CONNECTION'
            c2.conn_src_id = dest.id
            c2.conn_dest_id = src.id
            c2.conn_src_outlet = 0
            c2.conn_dest_inlet = len(src.pd_object.inlets()) - 1
            c2.connected = True

            self.nodes.append(c1)
            self.nodes.append(c2)

    def process_connection(self, c):
        if c.type == 'CONNECTION_MANUAL':
            c.connected = True
            return

        if c.type == 'CONNECTION_X':
            self.process_cross_connection(c)
            return

        if c.type == 'CONNECTION_LEFT':
            conn_start = [c.char_pos + c.tok.value.index('/')]
        else:
            conn_start = [c.char_pos]

        # find object on previous line
        src = self.find_connection(c.line_pos - 1, conn_start)
        if src is None:
            print("connection source is not found for: {0:s}".format(c))
            return

        c.conn_src_id = src.id
        if src.is_connection():
            # multiline connection
            c.conn_src_id = src.conn_src_id
            c.conn_src_outlet = src.conn_src_outlet
            src.connected = False

        # find on next line
        dest = self.find_connection(c.line_pos + 1, [c.char_pos, c.char_pos + c.width])
        if dest is None:
            print("connection destination is not found for: {0:s}".format(c.tok.value))
            return

        c.conn_dest_id = dest.id
        c.connected = True

    def layout_nodes(self):
        obj = filter(lambda x: x.is_object() and x.pd_object is not None, self.nodes)
        for n in obj:
            n.pd_object.x = n.char_pos * self.X_SPACE + self.X_PAD
            n.pd_object.y = n.line_pos * self.Y_SPACE + self.Y_PAD

    def num_lines(self):
        return len(self.lines)

    def num_elements(self, type):
        return len(self.elements(type))

    def elements(self, type):
        return filter(lambda x: x.type == type, self.nodes)

    def elements_in_line(self, type, line_pos):
        return filter(lambda x: x.type == type and x.line_pos == line_pos, self.nodes)

    def export(self, cnv):
        assert isinstance(cnv, Canvas)
        for n in filter(lambda x: x.is_object() and x.pd_object is not None, self.nodes):
            cnv.append_object(n.pd_object)

        for c in filter(lambda x: x.is_connection() and x.connected, self.nodes):
            # print("connection %s -> %s" % (c.conn_src_id, c.conn_dest_id))
            src = self.node_by_id(c.conn_src_id)
            dest = self.node_by_id(c.conn_dest_id)

            if src and dest:
                if not src.pd_object or not dest.pd_object:
                    logging.warning("can't connect {0:s} and {1:s}".format(src.tok, dest.tok))
                    return

                cnv.add_connection(
                    src.pd_object.id,
                    c.conn_src_outlet,
                    dest.pd_object.id,
                    c.conn_dest_inlet)



#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2014 by Serge Poltavski                                   #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

import six
import logging

from baseobject import BaseObject
from xletcalculator import XletCalculator
from xletpatchlookup import XletPatchLookup
from abstractvisitor import AbstractVisitor


class PdObject(BaseObject):
    xlet_calculator = XletCalculator()
    xlet_patch_finder = XletPatchLookup()
    # calculates xlets number
    XMETHOD_CALCULATE = 0
    # no calculation, xlets should be defined manually by calling set_inlets
    XMETHOD_EXPLICIT = 1
    # cache for external patches
    _patch_cache = {}
    # brect calc
    _brect_calc = None
    # fixed brect
    _fixed_size = False

    def __init__(self, name, x=0, y=0, w=0, h=0, args=None):
        BaseObject.__init__(self, x, y, w, h)
        if args is None:
            args = []
        assert isinstance(name, six.string_types)
        assert len(name) > 0

        self._name = name

        assert isinstance(args, list)
        self._args = args

        self._inlets = []
        self._outlets = []
        self._xlets_method = self.XMETHOD_CALCULATE
        self._gop = False

        if self.name in PdObject._patch_cache:
            if PdObject._patch_cache[self.name] is None:
                self._gop = False
            else:
                self._gop = True
                x, y, w, h = PdObject._patch_cache[self.name].gop_rect()
                self._width = w
                self._height = h
        else:
            # objname.pd patch found
            if self.xlet_patch_finder.has_object(self.name):
                import parser as pd

                obj = self.xlet_patch_finder.get_object(self.name)
                parser = pd.Parser()
                if not parser.parse(obj.path):
                    logging.error("can't parse patch: \"{0:s}\"".format(obj.path))
                    self._gop = False
                    PdObject._patch_cache[self.name] = None

                # success
                if parser.canvas.is_graph_on_parent():
                    self._gop = True
                    PdObject._patch_cache[self.name] = parser.canvas
                    x, y, w, h = parser.canvas.gop_rect()
                    self._width = w
                    self._height = h
                else:
                    self._gop = False
            else:
                PdObject._patch_cache[self.name] = None

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, n):
        assert isinstance(n, six.string_types)
        assert len(n) > 0
        self._name = n

    @property
    def args(self):
        return self._args

    def append_arg(self, a):
        self._args.append(a)

    @staticmethod
    def brect_calc():
        import brectcalculator

        if not PdObject._brect_calc:
            PdObject._brect_calc = brectcalculator.BRectCalculator()

        return PdObject._brect_calc

    @classmethod
    def add_object_xlet_info(cls, name, inlets, outlets):
        cls.xlet_calculator.mem_db.add_object(name, inlets, outlets)

    @classmethod
    def remove_object_xlet_info(cls, name):
        cls.xlet_calculator.mem_db.remove_object(name)

    def num_args(self):
        return len(self._args)

    def args_to_string(self):
        res = ""

        esc_args = []
        for arg in self.args:
            esc_args.append(PdObject.unescape(arg.encode('utf-8')))

        for arg in esc_args:
            if arg == ",":
                res += ", "
            elif arg == ";":
                res += ";"
            else:
                res += " " + arg

        res = " ".join(res.strip().split())
        return res

    def to_string(self):
        res = PdObject.unescape(self.name) + ' ' + self.args_to_string()
        return res.strip()

    def __str__(self):
        res = "[%-40s {x:%i,y:%i,id:%i}" % (self.to_string() + "]", self.x, self.y, self.id)
        return res

    def draw(self, painter):
        if self._gop:
            canvas = PdObject._patch_cache[self.name]
            painter.draw_gop(self, canvas)
        else:
            painter.draw_object(self)

    def inlets(self):
        if self._xlets_method == self.XMETHOD_EXPLICIT:
            return self._inlets
        elif self._xlets_method == self.XMETHOD_CALCULATE:
            return PdObject.xlet_calculator.inlets(self)
        else:
            return []

    def outlets(self):
        if self._xlets_method == self.XMETHOD_EXPLICIT:
            return self._outlets
        elif self._xlets_method == self.XMETHOD_CALCULATE:
            return PdObject.xlet_calculator.outlets(self)
        else:
            return []

    def traverse(self, visitor):
        assert isinstance(visitor, AbstractVisitor)

        if visitor.skip_object(self):
            return

        visitor.visit_object(self)

    def calc_brect(self, use_cached=True):
        if use_cached and (self._width != 0 and self._height != 0):
            return

        brect = self.brect_calc().object_brect(self)
        self._width = brect[2]
        self._height = brect[3]



#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2014 by Serge Poltavski                                 #
# serge.poltavski@gmail.com                                             #
# #
# This program is free software; you can redistribute it and/or modify  #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation; either version 3 of the License, or     #
# (at your option) any later version.                                   #
# #
# This program is distributed in the hope that it will be useful,       #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

import logging

from idocobjectvisitor import IDocObjectVisitor


def make_class_name(tag_name):
    return "Doc%s" % (tag_name.capitalize())


def create_instance(tag_name, *args):
    class_name = None
    try:
        class_name = make_class_name(tag_name)
        obj = globals()[class_name](args)
        return obj
    except KeyError:
        logging.error("Class not found: %s", class_name)
        exit(1)


class DocVisitorError(Exception):
    def __init__(self, message):
        self._msg = message

    def __str__(self):
        return "VisitorError: {0:s}".format(self._msg)


class DocItem(object):
    def __init__(self, *args):
        self._elements = []
        self._text = ""

        if len(args) > 1:
            self._text = args[0]

    def items(self):
        return self._elements

    def text(self):
        return self._text

    def is_valid_tag(self, tag_name):
        return False

    def add_child(self, element):
        self._elements.append(element)

    def clear(self):
        self._elements = []

    def num_children(self):
        return len(self._elements)

    def is_empty(self):
        return len(self._elements) == 0

    def traverse(self, visitor):
        assert isinstance(visitor, IDocObjectVisitor)

        try:
            mprefix = self.__class__.__name__[3:].lower()
            method = mprefix + "_begin"
            getattr(visitor, method)(self)
            # if hasattr(visitor, method):  # call element_begin() method
            #     getattr(visitor, method)(self)

            self.traverse_children(visitor)

            method = mprefix + "_end"
            # if hasattr(visitor, method):  # call element_end() method
            #     getattr(visitor, method)(self)
            getattr(visitor, method)(self)

        except DocVisitorError as e:
            logging.error(e)

    def traverse_children(self, visitor):
        for e in self._elements:
            e.traverse(visitor)

    def read_xml_data(self, xmlobj):
        self._text = xmlobj.text

    def from_xml(self, xmlobj):
        self.read_xml_data(xmlobj)

        for child in xmlobj:
            # skip XML comments
            if callable(child.tag):
                continue

            if not self.is_valid_tag(child.tag):
                logging.warning("tag <%s> not allowed in <%s>" % (child.tag, xmlobj.tag))
                continue
                # break

            obj = create_instance(child.tag)
            obj.from_xml(child)
            self.add_child(obj)


class DocTitle(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocMeta(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name in ("description", "authors", "contacts",
                            "license", "version", "website",
                            "aliases", "keywords", "library",
                            "also", "category", "since")


class DocDescription(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocAuthors(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name in ("author",)


class DocAuthor(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocAlso(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name == "see"


class DocSee(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._ref_view = "object"

    def from_xml(self, xmlobj):
        DocItem.from_xml(self, xmlobj)
        self._ref_view = xmlobj.get("view", "object")

    def is_link(self):
        return self._ref_view == "link"


class DocContacts(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocLibrary(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocVersion(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocCategory(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._ref_view = "object"

    def read_xml_data(self, xmlobj):
        DocItem.read_xml_data(self, xmlobj)
        self._ref_view = xmlobj.attrib.get("view", "object")

    def ref_view(self):
        return self._ref_view


class DocLicense(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._url = ""
        self._name = ""

    def read_xml_data(self, xmlobj):
        self._url = xmlobj.attrib.get("url", "")
        self._name = xmlobj.text

    def url(self):
        return self._url

    def name(self):
        return self._name


class DocExample(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name in ("pdexample", "pdinclude", "pdascii")


class DocPdobject(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._id = ""
        self._name = ""
        self._args = ""
        self._comment = ""
        self._offset = 0
        self._highlight = False
        self._attrs = {}

    @property
    def id(self):
        return self._id

    def is_valid_tag(self, tag_name):
        return tag_name in ("pdinlet", "pdoutlet")

    def highlight(self):
        return self._highlight == "true"

    def name(self):
        return self._name

    def offset(self):
        return self._offset

    def from_xml(self, xmlobj):
        try:
            self._id = xmlobj.attrib["id"]
            self._name = xmlobj.attrib["name"]
            self._args = xmlobj.attrib.get("args", "")
            self._comment = xmlobj.attrib.get("comment", "")
            self._offset = int(xmlobj.attrib.get("offset", "0"))
            self._highlight = xmlobj.attrib.get("highlight", "")

            # save other attrs
            for k, v in xmlobj.attrib.items():
                if k not in ('id', 'name', 'args', 'comment', 'offset', 'highlight'):
                    self._attrs[k] = v

            DocItem.from_xml(self, xmlobj)
        except KeyError as e:
            logging.warning("required attribute not found: \"%s\" in <%s>" % (e.message, xmlobj.tag))

    @property
    def comment(self):
        return self._comment

    def args(self):
        return self._args.split(" ")

    def attrs(self):
        return self._attrs


class DocPdmessage(DocPdobject):
    def __init__(self, *args):
        DocPdobject.__init__(self, args)

    def is_valid_tag(self, name):
        return False

    def from_xml(self, xmlobj):
        try:
            self._id = xmlobj.attrib["id"]
            self._comment = xmlobj.attrib.get("comment", "")
            self._offset = int(xmlobj.attrib.get("offset", "0"))
            DocItem.from_xml(self, xmlobj)
            # after parent method
            self._text = xmlobj.attrib["text"]
        except KeyError as e:
            logging.warning("required attribute not found: \"%s\" in <%s>" % (e.message, xmlobj.tag))


class DocPdinlet(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)

class DocPdinclude(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._file = None
        self._size = "auto"

    def file(self):
        return self._file

    def size(self):
        return self._size

    def from_xml(self, xmlobj):
        try:
            self._file = xmlobj.attrib['file']
            self._size = xmlobj.attrib.get('size', 'auto')
            assert self._size in ("auto", "canvas")
        except KeyError as e:
            logging.warning("attribute \"%s\" not found in <%s>", e.args[0], xmlobj.tag)


class DocPdascii(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._x_pad = 0
        self._y_pad = 0
        self._x_space = 0
        self._y_space = 0

    def x_pad(self):
        return self._x_pad

    def y_pad(self):
        return self._y_pad

    def x_space(self):
        return self._x_space

    def y_space(self):
        return self._y_space

    def from_xml(self, xmlobj):
        self.read_xml_data(xmlobj)
        try:
            self._x_pad = xmlobj.attrib.get('x-pad', 20)
            self._y_pad = xmlobj.attrib.get('y-pad', 20)
            self._x_space = xmlobj.attrib.get('x-space', 1.2)
            self._y_space = xmlobj.attrib.get('y-space', 1.2)
        except KeyError as e:
            logging.warning("attribute \"%s\" not found in <%s>", e.args[0], xmlobj.tag)


class DocPdoutlet(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocRow(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name in ("col", "pdmessage", "pdobject", "pdcomment")


class DocCol(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name in ("row", "pdmessage", "pdobject", "pdcomment")


class DocPdconnect(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._src_id = ""
        self._dest_id = ""
        self._src_out = None
        self._dest_in = None

    def src_id(self):
        return self._src_id

    def dest_id(self):
        return self._dest_id

    def from_xml(self, xmlobj):
        self._src_out = int(xmlobj.attrib["src_out"])
        self._dest_in = int(xmlobj.attrib["dest_in"])
        self._dest_id = xmlobj.attrib["dest"]
        self._src_id = xmlobj.attrib["src"]
        DocItem.from_xml(self, xmlobj)


class DocPdcomment(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocPdexample(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._width = 0
        self._height = 0
        self._file = ""
        self._size = ""
        self._title = ""

    def is_valid_tag(self, tag_name):
        return tag_name in ("row", "col", "pdconnect")

    def from_xml(self, xmlobj):
        size = xmlobj.attrib.get("size", None)
        self._file = xmlobj.attrib.get("file", None)
        if size:
            if size in ("auto", "canvas"):
                self._size = size
            else:
                logging.warning("invalid size value: \"{0:s}\" in <{1:s}>".format(size, xmlobj.tag))

        self._width = int(xmlobj.attrib.get("width", 0))
        self._height = int(xmlobj.attrib.get("height", 0))
        self._title = xmlobj.attrib.get("title", "")
        DocItem.from_xml(self, xmlobj)

    def file(self):
        return self._file

    def width(self):
        return self._width

    def height(self):
        return self._height

    def size(self):
        return self._size

    def title(self):
        return self._title


class DocWebsite(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocKeywords(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._keywords = []

    def from_xml(self, xmlobj):
        self._keywords = xmlobj.text.split(" ")
        super(self.__class__, self).from_xml(xmlobj)

    def keywords(self):
        return self._keywords


class DocAliases(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._aliases = []

    def aliases(self):
        return self._aliases

    def from_xml(self, xmlobj):
        for child in xmlobj:
            if child.tag == "alias":
                self._aliases.append(child.text)


class DocXlets(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._dynamic = False

    def pd_type_list(self):
        return map(lambda x: x.pd_type(), self.items())

    def enumerate(self):
        # enumerate
        for n in range(len(self.items())):
            self.items()[n].enumerate(n + 1)

    def is_dynamic(self):
        return self._dynamic

    def from_xml(self, xmlobj):
        DocItem.from_xml(self, xmlobj)
        self._dynamic = xmlobj.attrib.get("dynamic", False)


class DocInlets(DocXlets):
    def __init__(self, *args):
        DocXlets.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name == "inlet"


class DocTypeElement(DocItem):
    allowed_types = ("control", "audio", "gui")

    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._type = "control"

    def is_valid_type(self, t):
        return t in self.allowed_types

    def type(self):
        return self._type

    def from_xml(self, xmlobj):
        self._type = xmlobj.get("type", "control")
        DocItem.from_xml(self, xmlobj)

    def pd_type(self):
        if self._type == "control":
            return 0
        elif self._type == "audio":
            return 1
        elif self._type == "gui":
            return 2
        else:
            return -1


class DocXlet(DocTypeElement):
    def __init__(self, *args):
        DocTypeElement.__init__(self, *args)
        self._number = ""

    def from_xml(self, xmlobj):
        DocTypeElement.from_xml(self, xmlobj)
        self._number = xmlobj.get("number", "")

    def enumerate(self, n):
        if self._number == "":
            self._number = n

    def number(self):
        return self._number


class DocInlet(DocXlet):
    def __init__(self, *args):
        DocXlet.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name == "xinfo"


class DocXinfo(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._maxvalue = ""
        self._minvalue = ""
        self._on = ""

    def from_xml(self, xmlobj):
        self._maxvalue = xmlobj.attrib.get("maxvalue", "")
        self._minvalue = xmlobj.attrib.get("minvalue", "")
        self._on = xmlobj.attrib.get("on", "")
        DocItem.from_xml(self, xmlobj)

    def range(self):
        if not self._minvalue and not self._maxvalue:
            return ()

        return self.min(), self.max()

    def on(self):
        return self._on

    def min(self):
        return self._minvalue

    def max(self):
        return self._maxvalue


class DocOutlets(DocXlets):
    def __init__(self, *args):
        DocXlets.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name == "outlet"


class DocOutlet(DocXlet):
    def __init__(self, *args):
        DocXlet.__init__(self, args)


class DocArgument(DocItem):
    UNIT_MAP = {
        'herz': 'Hz',
        'kiloherz': 'KHz',
        'millisecond': 'ms',
        'second': 'sec',
        'decibel': 'db',
        'bpm': 'bpm',
        'percent': '%'
    }

    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._number = ""
        self._type = ""
        self._units = ""
        self._name = ""
        self._minvalue = ""
        self._maxvalue = ""
        self._default = ""
        self._required = False
        self._enum = []

    def from_xml(self, xmlobj):
        self._name = xmlobj.attrib.get("name", "anonym")
        self._number = xmlobj.attrib.get("number", "")
        self._type = xmlobj.attrib.get("type", "")
        self._units = xmlobj.attrib.get("units", "")
        self._minvalue = xmlobj.attrib.get("minvalue", "")
        self._maxvalue = xmlobj.attrib.get("maxvalue", "")
        self._default = xmlobj.attrib.get("default", "")
        self._required = xmlobj.attrib.get("required", False) == "true"
        enum_str = xmlobj.attrib.get("enum", "").strip()
        if len(enum_str) > 1:
            self._enum = enum_str.split(' ')

        DocItem.from_xml(self, xmlobj)

    def type(self):
        return self._type

    def units(self):
        if not self._units:
            return ""

        return self.UNIT_MAP[self._units]

    def name(self):
        return self._name

    def required(self):
        return self._required is True

    def optional(self):
        return self._required is False

    def default(self):
        return self._default

    def number(self):
        return self._number

    def enumerate(self, n):
        if self._number == "":
            self._number = n

    def type_info(self):
        if self._type == "float":
            if self._units:
                return self.units()
            else:
                return "*float*"
        else:
            return self._type

    def main_info_prefix(self):
        res = ""
        if self._name:
            res += "{0}".format(self._name)
        if self._units:
            res += "({0})".format(self.units())

        if res:
            res += ": "
        return res

    def main_info(self):
        res = self.main_info_prefix()
        res += self.text()
        return res.strip()

    def range(self):
        if not self._minvalue and not self._maxvalue:
            return ()

        return self.min(), self.max()

    def min(self):
        return self._minvalue

    def max(self):
        return self._maxvalue

    def enum(self):
        return self._enum


class DocArguments(DocXlets):
    def __init__(self, *args):
        DocXlets.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name == "argument"


class DocMethods(DocXlets):
    def __init__(self, *args):
        DocXlets.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name == "method"


class DocMethod(DocItem):
    def __init__(self, *args):
        self._name = ""
        DocItem.__init__(self, args)

    def name(self):
        return self._name

    def is_valid_tag(self, tag_name):
        return tag_name == "param"

    def from_xml(self, xmlobj):
        self._name = xmlobj.attrib.get("name", "")
        DocItem.from_xml(self, xmlobj)


class DocParam(DocArgument):
    def param_name(self):
        if self._units:
            return self.units().upper()
        elif self._name:
            name = self.name()
            if self.default():
                name += '=' + self.default()

            if self.optional():
                name = "[{0}]".format(name)

            return name
        else:
            return "X"


class DocInfo(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name in ("itemize", "par", "a")


class DocPar(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocA(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._url = None

    @property
    def url(self):
        return self._url

    def from_xml(self, xmlobj):
        self._url = xmlobj.attrib["href"]
        DocItem.from_xml(self, xmlobj)


class DocItemize(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name in ("item",)


class DocObject(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._name = ""
        self._type = "box"

    def is_valid_tag(self, tag_name):
        return tag_name in ("title", "meta", "inlets", "outlets",
                            "arguments", "properties", "info", "example", "methods")

    def from_xml(self, xobj):
        self._name = xobj.attrib["name"]
        self._type = xobj.attrib.get("type", "box")
        DocItem.from_xml(self, xobj)

    def name(self):
        return self._name

    def traverse(self, visitor):
        visitor.name = self._name
        DocItem.traverse(self, visitor)

    def is_box(self):
        return self._type == "box"

    def is_gui(self):
        return self._type == "gui"


class DocSince(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocProperties(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name == "property"


class DocProperty(DocArgument):
    def __init__(self, *args):
        DocArgument.__init__(self, args)
        self._readonly = False

    def from_xml(self, xmlobj):
        DocArgument.from_xml(self, xmlobj)
        self._readonly = xmlobj.attrib.get("readonly", "false") == "true"
        self._default = xmlobj.attrib.get("default", "")

    def readonly(self):
        return self._readonly

    def is_flag(self):
        return self._type == "flag"

    def is_alias(self):
        return self._type == "alias"



#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2015 by Serge Poltavski                                 #
# serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #


__author__ = 'Serge Poltavski'

import re
from pddoc.pd import XLET_MESSAGE, XLET_SOUND


def expr_args_parse(args):
    if len(args) == 0:
        return 1
    params = list(set(re.findall("(\$[fsi][0-9])", " ".join(args))))
    if len(params) == 0:
        return 1
    else:
        return int(max(params, key=lambda x: int(x[2]))[2])


def expr_tilde_args_parse(args):
    if len(args) == 0:
        return []
    inl = list(set(re.findall("(\$[fsiv][0-9])", " ".join(args))))
    if len(inl) == 0:
        return []
    else:
        inl.sort(key=lambda x: int(x[2:]))
        res = []
        for i in inl:
            if i == "$v1":
                continue
            if i[1] == 'v':
                res.append(XLET_SOUND)
            else:
                res.append(XLET_MESSAGE)
        return res


def func_list_in(args):
    if not args:
        return 2
    if args[0] in ("trim", "length"):
        return 1
    else:
        return 2


def func_list_out(args):
    if not args:
        return 1
    if args[0] == "split":
        return 3
    else:
        return 1


def fexpr_tilde_args_parse(args):
    if len(args) == 0:
        return []
    inl = list(set(re.findall("(\$[fsixy][0-9])", " ".join(args))))
    if len(inl) == 0:
        return []
    else:
        inl.sort(key=lambda x: int(x[2:]))
        res = []
        for i in inl:
            if i == "$x1":
                continue
            if i[1] == 'x':
                res.append(XLET_SOUND)
            else:
                res.append(XLET_MESSAGE)
        return res

_re_num = re.compile(r"^\d+$")

_objects = {
    "adc~": (
        lambda args: [XLET_MESSAGE],
        lambda args: (2 if len(args) == 0 else len(args)) * [XLET_SOUND]
    ),
    "dac~": (
        lambda args: (2 if len(args) == 0 else len(args)) * [XLET_SOUND],
        lambda args: []
    ),
    "readsf~": (
        lambda args: [XLET_MESSAGE],
        lambda args: (1 if len(args) == 0 else int(args[0])) * [XLET_SOUND] + [XLET_MESSAGE]
    ),
    "writesf~": (
        lambda args: (1 if len(args) == 0 else int(args[0])) * [XLET_SOUND],
        lambda args: []
    ),
    "expr": (
        lambda args: expr_args_parse(args) * [XLET_MESSAGE],
        lambda args: len(filter(None, " ".join(args).split(";"))) * [XLET_MESSAGE]
    ),
    "expr~": (
        lambda args: [XLET_SOUND] + expr_tilde_args_parse(args),
        lambda args: [XLET_SOUND]
    ),
    "fexpr~": (
        lambda args: [XLET_SOUND] + fexpr_tilde_args_parse(args),
        lambda args: [XLET_SOUND]
    ),
    "sprintf": (
        lambda args: (1 if not len(args) else (len(" ".join(args).replace("%%", "").split("%")) - 1)) * [
            XLET_MESSAGE],
        lambda args: (1 if " ".join(args).replace("%%", "").count("%") > 0 else 0) * [XLET_MESSAGE]
    ),
    "select": (
        lambda args: (2 if len(args) < 2 else 1) * [XLET_MESSAGE],
        lambda args: (2 if len(args) == 0 else len(args) + 1) * [XLET_MESSAGE]
    ),
    "sel": (
        lambda args: (2 if len(args) < 2 else 1) * [XLET_MESSAGE],
        lambda args: (2 if len(args) == 0 else len(args) + 1) * [XLET_MESSAGE]
    ),
    "route": (
        lambda args: (2 if len(args) < 2 else 1) * [XLET_MESSAGE],
        lambda args: (2 if len(args) == 0 else len(args) + 1) * [XLET_MESSAGE]
    ),
    "send": (
        lambda args: (2 if not args else 1) * [XLET_MESSAGE],
        lambda args: []
    ),
    "s": (
        lambda args: (2 if not args else 1) * [XLET_MESSAGE],
        lambda args: []
    ),
    "pointer": (
        lambda args: 2 * [XLET_MESSAGE],
        lambda args: (len(args) + 2) * [XLET_MESSAGE]
    ),
    "pack": (
        lambda args: (2 if len(args) == 0 else len(args)) * [XLET_MESSAGE],
        lambda args: [XLET_MESSAGE]
    ),
    "unpack": (
        lambda args: [XLET_MESSAGE],
        lambda args: (2 if len(args) == 0 else len(args)) * [XLET_MESSAGE]
    ),
    "trigger": (
        lambda args: [XLET_MESSAGE],
        lambda args: (2 if len(args) == 0 else len(args)) * [XLET_MESSAGE]
    ),
    "t": (
        lambda args: [XLET_MESSAGE],
        lambda args: (2 if len(args) == 0 else len(args)) * [XLET_MESSAGE]
    ),
    "notein": (
        lambda args: [],
        lambda args: (3 if len(args) == 0 else 2) * [XLET_MESSAGE]
    ),
    "ctlin": (
        lambda args: [],
        lambda args: (3 if len(args) == 0 else 2 if len(args) == 1 else 1) * [XLET_MESSAGE]
    ),
    "pgmin": (
        lambda args: [],
        lambda args: (2 if len(args) == 0 else 1) * [XLET_MESSAGE]
    ),
    "bendin": (
        lambda args: [],
        lambda args: (2 if len(args) == 0 else 1) * [XLET_MESSAGE]
    ),
    "touchin": (
        lambda args: [],
        lambda args: (2 if len(args) == 0 else 1) * [XLET_MESSAGE]
    ),
    "polytouchin": (
        lambda args: [],
        lambda args: (3 if len(args) == 0 else 2) * [XLET_MESSAGE]
    ),
    "list": (
        lambda args: func_list_in(args) * [XLET_MESSAGE],
        lambda args: func_list_out(args) * [XLET_MESSAGE]
    )
}


def is_snd_math(name):
    return name in ("+~", "*~", "-~", "/~", "max~", "min~")


def is_digit_object(name):
    return _re_num.search(name)


def inlets(name, args):
    # digits float object [2], [3] and others
    if is_digit_object(name):
        return [XLET_MESSAGE]

    if is_snd_math(name):

        if len(args) == 0:
            return [XLET_SOUND] * 2
        else:
            return [XLET_SOUND, XLET_MESSAGE]

    if name in _objects:
        return _objects[name][0](args)

    return []


def outlets(name, args):
    # digits float object [2], [3] and others
    if is_digit_object(name):
        return [XLET_MESSAGE] * 2

    if is_snd_math(name):
        return [XLET_SOUND]

    if name in _objects:
        return _objects[name][1](args)

    return []


def has_object(name):
    if is_snd_math(name) or is_digit_object(name):
        return True

    return name in _objects


#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2017 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

from __future__ import print_function
from pddoc.pd import PdObject
from pddoc.cairopainter import CairoPainter


class UIBase(PdObject):
    def __init__(self, name, x, y, **kwargs):
        PdObject.__init__(self, name, x, y, 100, 25, [])
        self._fixed_size = True
        self._properties = kwargs
        self.parse_props()

    def parse_props(self):
        if '@size' in self._properties:
            sz = self._properties['@size'].split('x')
            if len(sz) == 2:
                self.set_property('@size', sz[0] + ' ' + sz[1])
        else:
            self.set_property('@size', '100 25')

    def set_property(self, name, value):
        self._properties[name] = value
        if name == '@size':
            self.width = int(value.split(' ')[0])
            self.height = int(value.split(' ')[1])

    def get_property(self, name, default=None):
        return self._properties.get(name, default)

    @property
    def args(self):
        res = []
        for k, v in self._properties.items():
            res.append(k)
            if not v:
                continue

            for p in v.split(' '):
                res.append(str(p))

        return res

    def calc_brect(self):
        pass

    def draw(self, painter):
        assert isinstance(painter, CairoPainter)
        pass

    def set_bg_color(self, color):
        self.set_property('@bgcolor', ' '.join(map(lambda c: str(c), color.rgb_float())))




#!/usr/bin/env python
# coding=utf-8
#   Copyright (C) 2016 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

from lxml import etree
from pd.factory import make_by_name
from pd.obj import PdObject
from pdpage import PdPage


class LibraryParser(object):
    WINDOW_WIDTH = 715
    WINDOW_HEIGHT = 500
    HEADER_HEIGHT = 40
    OBJECT_OFFSET = 30
    DESCRIPTION_OFFSET = 175

    def __init__(self, fname):
        self._fname = fname
        self._xml = None
        self._root = None
        self._lib_name = ""
        self._lib_version = ""
        self._lib_authors = []
        self._lib_license = ""
        self._lib_website = ""
        self._lib_contacts = ""
        self._lib_info = ""
        self._current_y = self.HEADER_HEIGHT + 10
        self._pd_cats = {}
        self._pp = PdPage("lib", self.WINDOW_WIDTH, self.WINDOW_HEIGHT)

    def lib_name(self):
        return self._lib_name

    def process(self):
        self._xml = etree.parse(self._fname)
        self._xml.xinclude()
        self._root = self._xml.getroot()
        self.process_xml()
        self.add_header()
        self.add_footer()

    def process_xml(self):
        self.get_meta()
        # order matters! add_xlet_db() should called after get_meta()
        self.add_xlet_db()
        self.add_lib_description()
        self.process_xml_categories()

    def get_meta(self):
        self.get_lib_name()
        self.get_lib_version()
        self.get_lib_website()
        self.get_lib_license()

    def process_xml_categories(self):
        for c in self._root.findall("category"):
            self.process_xml_category(c)

    def process_xml_category(self, c):
        self.add_category_section(self._current_y, c.get('name'))
        self.process_xml_category_entries(c)

    def process_xml_category_entries(self, c):
        cat_name = c.get('name')
        for e in c.findall('entry'):
            self.add_object_description(e, cat_name)

    def add_xlet_db(self):
        PdObject.xlet_calculator.add_db(self._lib_name + ".db")

    def get_lib_name(self):
        self._lib_name = self._root.get("name")

    def get_lib_version(self):
        version = self._root.find("meta/version")
        if version is not None:
            self._lib_version = version.text
        else:
            self._lib_version = "0.0"

    def get_lib_website(self):
        el = self._root.find("meta/library-info/website")
        if el is not None:
            self._lib_website = el.text

    def get_lib_license(self):
        el = self._root.find("meta/library-info/license")
        if el is not None:
            self._lib_license = el.text

    def add_lib_description(self):
        lib_descr = self._root.find("meta/library-info/description")
        if lib_descr is not None:
            descr = self._pp.add_txt(lib_descr.text, self.OBJECT_OFFSET, self._current_y)
            self._current_y += descr.height + 10

    def add_object_description(self, obj, cat_name):
        pdobj = make_by_name(obj.get('name'))
        pdobj.x = self.OBJECT_OFFSET
        pdobj.y = self._current_y

        ref_view = obj.get('ref_view', 'object')
        if ref_view == 'object':
            self._pp.append_object(pdobj)
        elif ref_view == 'link':
            name = format(obj.get('name'))
            help_file = '{0}-help.pd'.format(name)
            link_text = '[{0}]'.format(name)
            self._pp.add_link(link_text, help_file, self.OBJECT_OFFSET, self._current_y)

        info = self._pp.add_txt(obj.get('descr'), self.DESCRIPTION_OFFSET, self._current_y)
        self._current_y += max(info.height, pdobj.height) + 10

        self._pd_cats[cat_name].append(pdobj)
        self._pd_cats[cat_name].append(info)

    def __str__(self):
        return self._pp.to_string()

    def add_header(self):
        title = "{0}".format(self._lib_name)
        h = self._pp.add_header(title)
        return h

    def add_footer(self):
        f = self._pp.make_footer(self._current_y)
        self._pp.append_object(f)

        info = self._pp.make_txt("version: v{0}, license: {1}".format(
            self._lib_version, self._lib_license), 20, f.top)
        self._pp.append_object(info)

        if self._lib_website:
            lnk = self._pp.make_link(20, f.top + info.height + 3, self._lib_website, self._lib_website)
            self._pp.append_object(lnk)
        return f

    def add_category_section(self, y, txt):
        l, r, brect = self._pp.add_section(txt, y)
        self._current_y = brect[1] + brect[3] + 10
        self._pd_cats[txt] = []



#!/usr/bin/env python
# coding=utf-8
# Copyright (C) 2014 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

import textwrap

from .message import Message
from .canvas import Canvas
from .abstractvisitor import AbstractVisitor


class BRectCalculator(AbstractVisitor):
    def __init__(self):
        from pddoc.cairopainter import CairoPainter
        self._cairo = CairoPainter(1, 1, None, "png")
        self._bboxes = []

    def clear(self):
        self._bboxes = []

    @property
    def bboxes(self):
        return self._bboxes

    def brect(self):
        left = min(self._bboxes, key=lambda o: o[0])[0]
        top = min(self._bboxes, key=lambda o: o[1])[1]
        right = max(self._bboxes, key=lambda o: o[0] + o[2])
        right = right[0] + right[2]
        bottom = max(self._bboxes, key=lambda o: o[1] + o[3])
        bottom = bottom[1] + bottom[3]

        return left, top, right - left, bottom - top

    def object_brect(self, obj):
        if obj._fixed_size:
            return obj.x, obj.y, obj.width, obj.height

        w, h = self._cairo.box_size(obj.to_string())
        return obj.x, obj.y, int(w), int(h)

    def text_brect(self, text):
        lines = []
        for line in textwrap.wrap(text, 59):
            line = ";\n".join(line.split(";")).split("\n")
            for subl in line:
                if subl:
                    lines.append(subl.strip())

        max_w = 0
        max_h = 0
        for line in lines:
            w, h = self._cairo.text_size(line)
            max_w = max(max_w, w)
            max_h = max(max_h, h)

        return 0, 0, int(max_w * 1.1), int(len(lines) * max_h)

    def comment_brect(self, comment):
        return self.text_brect(comment.text())

    def subpatch_brect(self, cnv):
        assert isinstance(cnv, Canvas)
        assert cnv.type == Canvas.TYPE_SUBPATCH

        txt = "pd " + cnv.args_to_string()
        w, h = self._cairo.box_size(txt)
        return cnv.x, cnv.y, int(w), int(h)

    def string_brect(self, string, font_size):
        w, h = self._cairo.text_size(string, font_size)
        return 0, 0, w, h

    def message_brect(self, message):
        assert isinstance(message, Message)
        w, h = self._cairo.message_size(message.to_string())
        return message.x, message.y, int(w), int(h)

    def visit_object(self, obj):
        bbox = self.object_brect(obj)
        obj.set_width(bbox[2])
        obj.set_height(bbox[3])
        self._bboxes.append(bbox)

    def visit_message(self, msg):
        bbox = self.message_brect(msg)
        msg.set_width(bbox[2])
        msg.set_height(bbox[3])
        self._bboxes.append(bbox)

    def visit_comment(self, comment):
        (x, y, w, h) = self.comment_brect(comment)
        self._bboxes.append((comment.x, comment.y, w, h))

    def visit_core_gui(self, gui):
        self._bboxes.append((gui.x, gui.y, gui.width, gui.height))

    def visit_graph(self, g):
        pass

    def visit_connection(self, c):
        pass

    def visit_canvas_end(self, canvas):
        pass

    def visit_canvas_begin(self, canvas):
        pass
